/**
 * Copyright (c) Members of the EGEE Collaboration. 2007-2010. 
 * See http://www.eu-egee.org/partners/ for details on the copyright
 * holders.  
 * 
 * Licensed under the Apache License, Version 2.0 (the "License"); 
 * you may not use this file except in compliance with the License. 
 * You may obtain a copy of the License at 
 * 
 *     http://www.apache.org/licenses/LICENSE-2.0 
 * 
 * Unless required by applicable law or agreed to in writing, software 
 * distributed under the License is distributed on an "AS IS" BASIS, 
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. 
 * See the License for the specific language governing permissions and 
 * limitations under the License.
 *
 *
 *  Authors:
 *  2007-
 *     Oscar Koeroo <okoeroo@nikhef.nl>
 *     David Groep <davidg@nikhef.nl>
 *     NIKHEF Amsterdam, the Netherlands
 *     <grid-mw-security@nikhef.nl> 
 *
 *
 */

#define _XOPEN_SOURCE	600

/* On Solaris need netdb.h with __EXTENSIONS__ for definition of MAXHOSTNAMELEN
 * */
#ifdef __sun
# define __EXTENSIONS__
#endif

#include "scas_config.h"

#include <stdlib.h>
#include <stdio.h>
#include <unistd.h>
#include <signal.h>
#include <string.h>
#include <sys/types.h>
#include <time.h>
#include <limits.h>
#include <sys/socket.h>
#include <netdb.h>
#include <errno.h>

#include <pthread.h>

#include "xacml_attr_obl_profile.h"
#include "pdp_xacml_authz_process.h"


#include "xacml_server.h"
#include "xacml_io_ssl.h"


#include "pdp_xacml_lcas_lcmaps.h"

#include "scas_log.h"


/* Solaris does not define HOST_NAME_MAX but MAXHOSTNAMELEN in netdb.h */
#ifdef __sun
# ifndef HOST_NAME_MAX
#  define HOST_NAME_MAX	MAXHOSTNAMELEN
# endif
#endif

/* BSDs also don't define HOST_NAME_MAX, fall back on _POSIX_HOST_NAME_MAX */
#ifndef HOST_NAME_MAX
# define HOST_NAME_MAX	_POSIX_HOST_NAME_MAX
#endif

/* Max length for issuer: https://<host>:<port>/ + \0 */
#define ISSUER_MAX			(HOST_NAME_MAX+15+1)


/************************************************************************
 * Global variables
 ************************************************************************/

/* Issuer used in response */
static char response_issuer[ISSUER_MAX]="SCAS";

/* Whether LCMAPS should use the cert_chain (when available) 
 * or the DN and FQANs */
static int use_cert_chain = 0;

#define	TIMESTRINGLEN	21L


/************************************************************************
 * Private prototypes
 ************************************************************************/

#ifdef HAVE_XACML_REQUEST_GET_ID
/* Converts a time_t to a char* */
static int time_t_to_gmtime_str(const time_t t, char *buf, size_t buf_len);
#endif /* HAVE_XACML_REQUEST_GET_ID */

/************************************************************************
 * Public functions
 ************************************************************************/

/************************************************
Function:    extractRequestSubject
Parameters:  const xacml_request_t           request
             xacml_profile_request_t * xacml_profile_request
Description:
             This function extracts the Subject information details
             from the request and puts that into the xacml_profile_request_t 
             object.
Return:
          0 : good
         !0 : bad 
************************************************/
int extractRequestSubject (const xacml_request_t           request,
                                 xacml_profile_request_t * xacml_profile_request)
{
    const char * logstr = "extractRequestSubject()";
    size_t count = 0;
    size_t i     = 0;

    xacml_request_get_subject_attribute_count(request, &count);

    for (i = 0; i < count; i++)
    {
        const char *category;
        const char *attribute_id;
        const char *data_type;
        const char *issuer;
        const char *value;

        xacml_request_get_subject_attribute(
                request, i, &category, &attribute_id, &data_type, &issuer, &value);

        /* Should only be send once */
        if (strcmp (attribute_id, XACML_ATTR_PROFILE_NS_URL_SUBJECT_X509_ID) == 0)                    
        {
            if (!(xacml_profile_request -> xacml_profile_subject.xacml_subject_x509_id.subject_x509_id                             = strdup (value)))
            { 
                scas_log(LOG_ERR, "%s: Failure in allocating memory @  1\n", logstr);
                return -1;
            }
            continue;
        }
        if (strcmp (attribute_id, XACML_ATTR_PROFILE_NS_URL_SUBJECT_X509_ISSUER) == 0)                
        {
            if (!(xacml_profile_request -> xacml_profile_subject.xacml_subject_x509_issuer.subject_x509_issuer                     = strdup (value)))
            { 
                scas_log(LOG_ERR, "%s: Failure in allocating memory @  2\n", logstr);
                return -1;
            }
            continue;
        }
        if (strcmp (attribute_id, XACML_ATTR_PROFILE_NS_URL_SUBJECT_CERTIFICATE_SERIAL_NUMBER) == 0)  
        {
            xacml_profile_request -> xacml_profile_subject.xacml_subject_certificate_serial_number.certificate_serial_number = atoi (value);
            /* Should handle atoi() failures */
            continue;
        }
        if (strcmp (attribute_id, XACML_ATTR_PROFILE_NS_URL_CERT_CHAIN) == 0)
        {
            if (!(xacml_profile_request -> xacml_profile_subject.xacml_cert_chain.cert_chain                                       = strdup (value)))
            { 
                scas_log(LOG_ERR, "%s: Failure in allocating memory @  3\n", logstr);
                return -1;
            }
            continue;
        }
        if (strcmp (attribute_id, XACML_ATTR_PROFILE_NS_URL_SUBJECT_VOMS_PRIMARY_FQAN) == 0)
        {
            if (!(xacml_profile_request -> xacml_profile_subject.xacml_subject_primary_voms_fqan.subject_primary_voms_fqan         = strdup (value)))
            { 
                scas_log(LOG_ERR, "%s: Failure in allocating memory @  4\n", logstr);
                return -1;
            }
            continue;
        }


        /* Multiple elements */
        if (strcmp (attribute_id, XACML_ATTR_PROFILE_NS_URL_SUBJECT_CA_SERIAL_NUMBER) == 0)
        {
            xacml_profile_request -> xacml_profile_subject.xacml_ca_serial_number_cnt++;
            if (!(xacml_profile_request -> xacml_profile_subject.xacml_ca_serial_number = realloc (xacml_profile_request -> xacml_profile_subject.xacml_ca_serial_number, 
                                                                                                   sizeof (xacml_ca_serial_number_t) * (xacml_profile_request -> xacml_profile_subject.xacml_ca_serial_number_cnt))))
            {
                /* Restore of the counter, needed for the clean up function */
                xacml_profile_request -> xacml_profile_subject.xacml_ca_serial_number_cnt--;
                scas_log(LOG_ERR, "%s: Failure in allocating memory @  5\n", logstr);
                return -1;
            }

            xacml_profile_request -> xacml_profile_subject.xacml_ca_serial_number[xacml_profile_request -> xacml_profile_subject.xacml_ca_serial_number_cnt - 1].ca_serial_number = atoi(value);
            continue;
        }
        if (strcmp (attribute_id, XACML_ATTR_PROFILE_NS_URL_CA_POLICY_OID) == 0)
        {
            xacml_profile_request -> xacml_profile_subject.xacml_ca_policy_oid_cnt++;
            if (!(xacml_profile_request -> xacml_profile_subject.xacml_ca_policy_oid = realloc (xacml_profile_request -> xacml_profile_subject.xacml_ca_policy_oid, 
                                                                                                sizeof (xacml_ca_policy_oid_t) * xacml_profile_request -> xacml_profile_subject.xacml_ca_policy_oid_cnt)))
            {
                /* Restore of the counter, needed for the clean up function */
                xacml_profile_request -> xacml_profile_subject.xacml_ca_policy_oid_cnt--;   
                scas_log(LOG_ERR, "%s: Failure in allocating memory @  6\n", logstr);
                return -1;
            }

            if (!(xacml_profile_request -> xacml_profile_subject.xacml_ca_policy_oid[xacml_profile_request -> xacml_profile_subject.xacml_ca_policy_oid_cnt - 1].ca_policy_oid          = strdup (value)))
            {
                scas_log(LOG_ERR, "%s: Failure in allocating memory @  7\n", logstr);
                return -1;
            }
            continue;
        }
        if (strcmp (attribute_id, XACML_ATTR_PROFILE_NS_URL_VO) == 0)
        {
            xacml_profile_request -> xacml_profile_subject.xacml_subject_vo_cnt++;
            if (!(xacml_profile_request -> xacml_profile_subject.xacml_subject_vo = realloc (xacml_profile_request -> xacml_profile_subject.xacml_subject_vo, 
                                                                                             sizeof (xacml_subject_vo_t) * xacml_profile_request -> xacml_profile_subject.xacml_subject_vo_cnt)))
            {
                scas_log(LOG_ERR, "%s: Failure in allocating memory @  8\n", logstr);
                /* Restore of the counter, needed for the clean up function */
                xacml_profile_request -> xacml_profile_subject.xacml_subject_vo_cnt--;
                return -1;
            }

            if (!(xacml_profile_request -> xacml_profile_subject.xacml_subject_vo[xacml_profile_request -> xacml_profile_subject.xacml_subject_vo_cnt - 1].subject_vo = strdup (value)))
            {
                scas_log(LOG_ERR, "%s: Failure in allocating memory @  9\n", logstr);
                return -1;
            }
            continue;
        }
        if (strcmp (attribute_id, XACML_ATTR_PROFILE_NS_URL_VOMS_SIGNING_SUBJECT) == 0)
        {
            xacml_profile_request -> xacml_profile_subject.xacml_voms_signing_subject_cnt++;
            if (!(xacml_profile_request -> xacml_profile_subject.xacml_voms_signing_subject = realloc (xacml_profile_request -> xacml_profile_subject.xacml_voms_signing_subject, 
                                                                                                       sizeof (xacml_voms_signing_subject_t) * xacml_profile_request -> xacml_profile_subject.xacml_voms_signing_subject_cnt)))
            {
                /* Restore of the counter, needed for the clean up function */
                scas_log(LOG_ERR, "%s: Failure in allocating memory @ 10\n", logstr);
                xacml_profile_request -> xacml_profile_subject.xacml_voms_signing_subject_cnt--;
                return -1;
            }

            if (!(xacml_profile_request -> xacml_profile_subject.xacml_voms_signing_subject[xacml_profile_request -> xacml_profile_subject.xacml_voms_signing_subject_cnt - 1].voms_signing_subject = strdup (value)))
            {
                scas_log(LOG_ERR, "%s: Failure in allocating memory @ 11\n", logstr);
                return -1;
            }
            continue;
        }
        if (strcmp (attribute_id, XACML_ATTR_PROFILE_NS_URL_VOMS_SIGNING_ISSUER) == 0)
        {
            xacml_profile_request -> xacml_profile_subject.xacml_voms_signing_issuer_cnt++;
            if (!(xacml_profile_request -> xacml_profile_subject.xacml_voms_signing_issuer = realloc (xacml_profile_request -> xacml_profile_subject.xacml_voms_signing_issuer, 
                                                                                                      sizeof (xacml_voms_signing_issuer_t) * xacml_profile_request -> xacml_profile_subject.xacml_voms_signing_issuer_cnt)))
            {
                /* Restore of the counter, needed for the clean up function */
                xacml_profile_request -> xacml_profile_subject.xacml_voms_signing_issuer_cnt--;
                scas_log(LOG_ERR, "%s: Failure in allocating memory @ 12\n", logstr);
                return -1;
            }

            if (!(xacml_profile_request -> xacml_profile_subject.xacml_voms_signing_issuer[xacml_profile_request -> xacml_profile_subject.xacml_voms_signing_issuer_cnt - 1].voms_signing_issuer = strdup (value)))
            {
                scas_log(LOG_ERR, "%s: Failure in allocating memory @ 13\n", logstr);
                return -1;
            }
            continue;
        }
        if (strcmp (attribute_id, XACML_ATTR_PROFILE_NS_URL_VOMS_DNS_PORT) == 0)
        {
            xacml_profile_request -> xacml_profile_subject.xacml_voms_dns_port_cnt++;
            if (!(xacml_profile_request -> xacml_profile_subject.xacml_voms_dns_port = realloc (xacml_profile_request -> xacml_profile_subject.xacml_voms_dns_port, 
                                                                                                sizeof (xacml_voms_dns_port_t) * xacml_profile_request -> xacml_profile_subject.xacml_voms_dns_port_cnt)))
            {
                /* Restore of the counter, needed for the clean up function */
                xacml_profile_request -> xacml_profile_subject.xacml_voms_dns_port_cnt--;
                scas_log(LOG_ERR, "%s: Failure in allocating memory @ 14\n", logstr);
                return -1;
            }

            if (!(xacml_profile_request -> xacml_profile_subject.xacml_voms_dns_port[xacml_profile_request -> xacml_profile_subject.xacml_voms_dns_port_cnt - 1].voms_dns_port = strdup (value)))
            {
                scas_log(LOG_ERR, "%s: Failure in allocating memory @ 15\n", logstr);
                return -1;
            }  
            continue;
        }
        if (strcmp (attribute_id, XACML_ATTR_PROFILE_NS_URL_SUBJECT_VOMS_FQAN) == 0)
        {
            xacml_profile_request -> xacml_profile_subject.xacml_subject_voms_fqan_cnt++;
            if (!(xacml_profile_request -> xacml_profile_subject.xacml_subject_voms_fqan = realloc (xacml_profile_request -> xacml_profile_subject.xacml_subject_voms_fqan, 
                                                                                                    sizeof (xacml_subject_voms_fqan_t) * xacml_profile_request -> xacml_profile_subject.xacml_subject_voms_fqan_cnt)))
            
            {
                /* Restore of the counter, needed for the clean up function */
                xacml_profile_request -> xacml_profile_subject.xacml_subject_voms_fqan_cnt--;
                scas_log(LOG_ERR, "%s: Failure in allocating memory @ 16\n", logstr);
                return -1;
            }  

            if (!(xacml_profile_request -> xacml_profile_subject.xacml_subject_voms_fqan[xacml_profile_request -> xacml_profile_subject.xacml_subject_voms_fqan_cnt - 1].subject_voms_fqan = strdup (value)))
            {
                scas_log(LOG_ERR, "%s: Failure in allocating memory @ 17\n", logstr);
                return -1;
            }  
            continue;
        }



        if (strcmp(category, XACML_SUBJECT_CATEGORY_ACCESS_SUBJECT))
        {
            continue;
        }
        if (strcmp(attribute_id, XACML_SUBJECT_ATTRIBUTE_SUBJECT_ID))
        {
            continue;
        }
    }

    return 0;
}



/************************************************
Function:    extractRequestAction
Parameters:  const xacml_request_t           request
             xacml_profile_request_t * xacml_profile_request
Description:
             This function extracts the Action information details
             from the request and puts that into the xacml_profile_request_t 
             object.
Return:
          0 : good
         !0 : bad 
************************************************/
int extractRequestAction (const xacml_request_t           request,
                                xacml_profile_request_t * xacml_profile_request)
{
    xacml_result_t                      rc;
    size_t                              i;
    size_t                              count;
    const char *                        attribute_id;
    const char *                        data_type;
    const char *                        issuer;
    const char *                        value;
    const char *                        logstr = "extractRequestAction";

    rc = xacml_request_get_action_attribute_count(request, &count);
    if (rc!=XACML_RESULT_SUCCESS)	{
	scas_log(LOG_ERR, "%s: xacml_request_get_action_attribute_count() failed\n",logstr);
	return -1;
    }


    for (i = 0 ; i < count; i++)
    {
        rc = xacml_request_get_action_attribute(
                request,
                i,
                &attribute_id,
                &data_type,
                &issuer,
                &value);
	if (rc!=XACML_RESULT_SUCCESS)	{
	    scas_log(LOG_ERR, "%s: xacml_request_get_action_attribute() failed\n",logstr);
	    return -1;
	}

        if (strcmp (attribute_id, XACML_ATTR_PROFILE_NS_URL_ACTION_ACTION_ID) == 0)
        {
            if (strcmp (value, XACML_ATTR_PROFILE_NS_URL_ACTION_ACTION_TYPE_ACCESS_FILE) == 0)
            {
                xacml_profile_request -> xacml_profile_action.xacml_action_job_type  = ACTION_ACCESSFILE;

            }
            else if (strcmp (value, XACML_ATTR_PROFILE_NS_URL_ACTION_ACTION_TYPE_QUEUE) == 0) 
            {
                xacml_profile_request -> xacml_profile_action.xacml_action_job_type = ACTION_QUEUE;

            }
            else if (strcmp (value, XACML_ATTR_PROFILE_NS_URL_ACTION_ACTION_TYPE_EXEC_NOW) == 0) 
            {
                xacml_profile_request -> xacml_profile_action.xacml_action_job_type = ACTION_EXECUTE_NOW;

            }
            else
            { 
                scas_log(LOG_NOTICE, "%s: Unsupported Action-id '%s' found: %s\n", logstr, attribute_id, value);
            }
        }
        else if (strcmp (attribute_id, XACML_ATTR_PROFILE_NS_URL_ACTION_ACTION_RSL_STRING) == 0)
        {
            if (!(xacml_profile_request -> xacml_profile_action.xacml_action_rsl_string.action_rsl_string = strdup(value)))
            {
                scas_log(LOG_ERR, "%s: Failure in allocating memory @ 1\n", logstr);
                return -1;
            } 
        }
        else
        { 
            scas_log(LOG_DEBUG, "%s: Unknown action attribute-id '%s' found with value %s\n", logstr, attribute_id, value);
        }
    }

    return 0;
}



/************************************************
Function:    extractRequestResource
Parameters:  const xacml_request_t           request
             xacml_profile_request_t * xacml_profile_request
Description:
             This function extracts the Action information details
             from the request and puts that into the xacml_profile_request_t 
             object.
Return:
          0 : good
         !0 : bad 
************************************************/
int extractRequestResource (const xacml_request_t           request,
                                  xacml_profile_request_t * xacml_profile_request)
{
    xacml_result_t                      rc;
    size_t                              i;
    size_t                              j;
    size_t                              count;
    const char *                        attribute_id;
    const char *                        data_type;
    const char *                        issuer;
    const char *                        value;
    xacml_resource_attribute_t          resource_attribute  = NULL;
    size_t                              resource_attr_count = 0;
    const char *                        logstr = "extractRequestResource";


    rc = xacml_request_get_resource_attribute_count(request, &resource_attr_count);
    if (rc!=XACML_RESULT_SUCCESS)	{
	scas_log(LOG_ERR, "%s: xacml_request_get_resource_attribute_count() failed\n",logstr);
	return -1;
    }

    for (i = 0 ; i < resource_attr_count; i++)
    {
        rc = xacml_request_get_resource_attribute(
                request,
                i,
                &resource_attribute);
	if (rc!=XACML_RESULT_SUCCESS)	{
	    scas_log(LOG_ERR, "%s: xacml_request_get_resource_attribute() failed\n",logstr);
	    return -1;
	}


        if (resource_attribute)
        {
            rc = xacml_resource_attribute_get_count(
                       resource_attribute,
                       &count);
	    if (rc!=XACML_RESULT_SUCCESS)	{
		scas_log(LOG_ERR, "%s: xacml_resource_attribute_get_count() failed\n",logstr);
		return -1;
	    }

            for (j = 0; j < count; j++)
            {
                rc = xacml_resource_attribute_get_attribute(
                           resource_attribute,
                           j,
                           &attribute_id,
                           &data_type,
                           &issuer,
                           &value);
		if (rc!=XACML_RESULT_SUCCESS)	{
		    scas_log(LOG_ERR, "%s: xacml_resource_attribute_get_attribute() failed\n",logstr);
		    return -1;
		}

                if (strcmp (attribute_id, XACML_ATTR_PROFILE_NS_URL_RESOURCE_RESOURCE_ID) == 0)
                {
                    if (strcmp (value, XACML_ATTR_PROFILE_NS_URL_RESOURCE_RESOURCE_TYPE_CE) == 0)
                    {
                        xacml_profile_request -> xacml_profile_resource.xacml_resource_node_type = RESOURCE_CE;

                    }
                    else if (strcmp (value, XACML_ATTR_PROFILE_NS_URL_RESOURCE_RESOURCE_TYPE_SE) == 0)
                    {
                        xacml_profile_request -> xacml_profile_resource.xacml_resource_node_type = RESOURCE_SE;

                    }
                    else if (strcmp (value, XACML_ATTR_PROFILE_NS_URL_RESOURCE_RESOURCE_TYPE_WN) == 0)
                    {
                        xacml_profile_request -> xacml_profile_resource.xacml_resource_node_type = RESOURCE_WN;

                    }
                    else if (strcmp (value, XACML_ATTR_PROFILE_NS_URL_RESOURCE_RESOURCE_TYPE_RB) == 0)
                    {
                        xacml_profile_request -> xacml_profile_resource.xacml_resource_node_type = RESOURCE_RB;

                    }
                }
                else if (strcmp (attribute_id, XACML_ATTR_PROFILE_NS_URL_RESOURCE_RESOURCE_X509_ID) == 0)
                {
                    if (!(xacml_profile_request -> xacml_profile_resource.xacml_resource_x509_id.resource_x509_id = strdup(value)))
                    {
                        scas_log(LOG_ERR, "%s: Failure in allocating memory @ 1\n", logstr);
                        return -1;
                    } 
                }
                else if (strcmp (attribute_id, XACML_ATTR_PROFILE_NS_URL_RESOURCE_RESOURCE_X509_ISSUER) == 0)
                {
                    if (!(xacml_profile_request -> xacml_profile_resource.xacml_resource_x509_issuer.resource_x509_issuer = strdup(value)))
                    {
                        scas_log(LOG_ERR, "%s: Failure in allocating memory @ 2\n", logstr);
                        return -1;
                    } 
                }
                else if (strcmp (attribute_id, XACML_ATTR_PROFILE_NS_URL_RESOURCE_DNS_HOST_NAME) == 0)
                {
                    if (!(xacml_profile_request -> xacml_profile_resource.xacml_resource_dns_name.resource_dns_name = strdup(value)))
                    {
                        scas_log(LOG_ERR, "%s: Failure in allocating memory @ 3\n", logstr);
                        return -1;
                    } 
                }
                else
                { 
                    scas_log(LOG_DEBUG, "%s: Unknown resource attribute-id '%s' found with value %s\n", logstr, attribute_id, value);
                }
            }
        }
    }
    return 0;
}



/************************************************
Function:    destroyRequestSubject
Parameters:  xacml_profile_request_t * xacml_profile_request
Description:
             This function will destroy and free the Subject information details
             from the request, which was set by the extractRequestSubject() function
             to a xacml_profile_request_t
Return:
          0 : good
         !0 : bad 
************************************************/
int destroyRequestSubject (xacml_profile_request_t * xacml_profile_request)
{
    size_t i     = 0;


    if (xacml_profile_request -> xacml_profile_subject.xacml_subject_x509_id.subject_x509_id)
    {
        free(xacml_profile_request -> xacml_profile_subject.xacml_subject_x509_id.subject_x509_id);
        xacml_profile_request -> xacml_profile_subject.xacml_subject_x509_id.subject_x509_id = NULL;
    }
    if (xacml_profile_request -> xacml_profile_subject.xacml_subject_x509_issuer.subject_x509_issuer)
    {
        free(xacml_profile_request -> xacml_profile_subject.xacml_subject_x509_issuer.subject_x509_issuer);
        xacml_profile_request -> xacml_profile_subject.xacml_subject_x509_issuer.subject_x509_issuer = NULL; 
    } 
    if (xacml_profile_request -> xacml_profile_subject.xacml_cert_chain.cert_chain)
    {
        free(xacml_profile_request -> xacml_profile_subject.xacml_cert_chain.cert_chain);
        xacml_profile_request -> xacml_profile_subject.xacml_cert_chain.cert_chain = NULL;
    }

    if (xacml_profile_request -> xacml_profile_subject.xacml_subject_primary_voms_fqan.subject_primary_voms_fqan)
    {
        free(xacml_profile_request -> xacml_profile_subject.xacml_subject_primary_voms_fqan.subject_primary_voms_fqan);
        xacml_profile_request -> xacml_profile_subject.xacml_subject_primary_voms_fqan.subject_primary_voms_fqan = NULL;
    }

    /* The CA_serial number can appear multiple times, but is an integer which doesn't need the loop-ed free() construct */ 
    free(xacml_profile_request -> xacml_profile_subject.xacml_ca_serial_number);
    xacml_profile_request -> xacml_profile_subject.xacml_ca_serial_number = NULL;

    /* Freeing multi-valued elements */
    for (i = 0; i < xacml_profile_request -> xacml_profile_subject.xacml_ca_policy_oid_cnt; i++)
    {
        free(xacml_profile_request -> xacml_profile_subject.xacml_ca_policy_oid[i].ca_policy_oid);
        xacml_profile_request -> xacml_profile_subject.xacml_ca_policy_oid[i].ca_policy_oid = NULL;
    }
    if (xacml_profile_request -> xacml_profile_subject.xacml_ca_policy_oid)
    {
        free(xacml_profile_request -> xacml_profile_subject.xacml_ca_policy_oid);
        xacml_profile_request -> xacml_profile_subject.xacml_ca_policy_oid = NULL;
    }


    for (i = 0; i < xacml_profile_request -> xacml_profile_subject.xacml_subject_vo_cnt; i++)
    {
        free(xacml_profile_request -> xacml_profile_subject.xacml_subject_vo[i].subject_vo);
        xacml_profile_request -> xacml_profile_subject.xacml_subject_vo[i].subject_vo = NULL; 
    } 
    if (xacml_profile_request -> xacml_profile_subject.xacml_subject_vo)
    {
        free(xacml_profile_request -> xacml_profile_subject.xacml_subject_vo);
        xacml_profile_request -> xacml_profile_subject.xacml_subject_vo = NULL;
    }

    
    for (i = 0; i < xacml_profile_request -> xacml_profile_subject.xacml_voms_signing_subject_cnt; i++)
    {
        free(xacml_profile_request -> xacml_profile_subject.xacml_voms_signing_subject[i].voms_signing_subject);
        xacml_profile_request -> xacml_profile_subject.xacml_voms_signing_subject[i].voms_signing_subject = NULL;
    }
    if (xacml_profile_request -> xacml_profile_subject.xacml_voms_signing_subject)
    {
        free(xacml_profile_request -> xacml_profile_subject.xacml_voms_signing_subject);
        xacml_profile_request -> xacml_profile_subject.xacml_voms_signing_subject = NULL;
    }


    for (i = 0; i < xacml_profile_request -> xacml_profile_subject.xacml_voms_signing_issuer_cnt; i++)
    {
        free(xacml_profile_request -> xacml_profile_subject.xacml_voms_signing_issuer[i].voms_signing_issuer);
        xacml_profile_request -> xacml_profile_subject.xacml_voms_signing_issuer[i].voms_signing_issuer = NULL;
    }
    if (xacml_profile_request -> xacml_profile_subject.xacml_voms_signing_issuer)
    {
        free(xacml_profile_request -> xacml_profile_subject.xacml_voms_signing_issuer);
        xacml_profile_request -> xacml_profile_subject.xacml_voms_signing_issuer = NULL;
    }


    for (i = 0; i < xacml_profile_request -> xacml_profile_subject.xacml_voms_dns_port_cnt; i++)
    {
        free(xacml_profile_request -> xacml_profile_subject.xacml_voms_dns_port[i].voms_dns_port);
        xacml_profile_request -> xacml_profile_subject.xacml_voms_dns_port[i].voms_dns_port = NULL;
    }
    if (xacml_profile_request -> xacml_profile_subject.xacml_voms_dns_port)
    {
        free(xacml_profile_request -> xacml_profile_subject.xacml_voms_dns_port);
        xacml_profile_request -> xacml_profile_subject.xacml_voms_dns_port = NULL;
    }

    for (i = 0; i < xacml_profile_request -> xacml_profile_subject.xacml_subject_voms_fqan_cnt; i++)
    {
        free(xacml_profile_request -> xacml_profile_subject.xacml_subject_voms_fqan[i].subject_voms_fqan);
        xacml_profile_request -> xacml_profile_subject.xacml_subject_voms_fqan[i].subject_voms_fqan = NULL;
    }
    if (xacml_profile_request -> xacml_profile_subject.xacml_subject_voms_fqan)
    {
        free(xacml_profile_request -> xacml_profile_subject.xacml_subject_voms_fqan);
        xacml_profile_request -> xacml_profile_subject.xacml_subject_voms_fqan = NULL;
    }


    return 0;
}


/************************************************
Function:    destroyRequestAction
Parameters:  xacml_profile_request_t * xacml_profile_request
Description:
             This function will destroy and free the Action information details
             from the request, which was set by the extractRequestAction() function
             to a xacml_profile_request_t
Return:
          0 : good
         !0 : bad 
************************************************/
int destroyRequestAction (xacml_profile_request_t * xacml_profile_request)
{
    if (xacml_profile_request -> xacml_profile_action.xacml_action_rsl_string.action_rsl_string)
    {
        free (xacml_profile_request -> xacml_profile_action.xacml_action_rsl_string.action_rsl_string);
    }

    return 0;
}



/************************************************
Function:    destroyRequestResource
Parameters:  xacml_profile_request_t * xacml_profile_request
Description:
             This function will destroy and free the Resource information details
             from the request, which was set by the extractRequestResource() function
             to a xacml_profile_request_t
Return:
          0 : good
         !0 : bad 
************************************************/
int destroyRequestResource (xacml_profile_request_t * xacml_profile_request)
{
    if (xacml_profile_request -> xacml_profile_resource.xacml_resource_x509_id.resource_x509_id)
    {
        free (xacml_profile_request -> xacml_profile_resource.xacml_resource_x509_id.resource_x509_id);
    }
    if (xacml_profile_request -> xacml_profile_resource.xacml_resource_x509_issuer.resource_x509_issuer)
    {
        free (xacml_profile_request -> xacml_profile_resource.xacml_resource_x509_issuer.resource_x509_issuer);
    }
    if (xacml_profile_request -> xacml_profile_resource.xacml_resource_dns_name.resource_dns_name)
    {
        free (xacml_profile_request -> xacml_profile_resource.xacml_resource_dns_name.resource_dns_name);
    }

    return 0;
}


/************************************************
Function:    constructFailureResponse
Parameters:  xacml_response_t * response
Description:
             The XACML response message to indicate a failure is constructed here.
************************************************/
int constructFailureResponse (xacml_response_t * response)
{
    xacml_response_set_saml_status_code  (*response, SAML_STATUS_AuthnFailed);
    xacml_response_set_xacml_status_code (*response, XACML_STATUS_ok);
    xacml_response_set_xacml_decision    (*response, XACML_DECISION_Deny);

    return 0;
}



/************************************************
Function:    constructResponse
Parameters:  xacml_response_t * response
             uid_t uid
             gid_t pgid
             gid_t sgids[]
             int   sgid_cnt
Description:
             The XACML response message is constructed here.
             This implementation will use the Unix UID, Primary GID and
             multiple Secondary GIDs as input to contruct the obligations and
             its attributes. 
************************************************/
int constructResponse (xacml_response_t * response,
                       uid_t uid,
                       gid_t pgid,
                       gid_t sgids[],
                       int   sgid_cnt)
{
    /* Mapping Information Translated */
    char *  str_uid       = NULL;
    char *  str_pgid      = NULL;
    char ** str_sgids     = NULL;
    size_t  str_sgids_cnt = 0;
    size_t  i             = 0;
    /* Mapping Information Translated */

    xacml_obligation_t  obligation_uidgid                = NULL;
    xacml_obligation_t  obligation_secgids               = NULL;


    /* transform! - 2^32 uses 10 bytes */
    /* str_uid = malloc (sizeof (char) * 11); */
    str_uid = calloc (11, sizeof (char));
    snprintf (str_uid, 11, "%u", (unsigned)uid);

    /* str_pgid = malloc (sizeof (char) * 11); */
    str_pgid = calloc (11, sizeof (char));
    snprintf (str_pgid, 11, "%u", (unsigned)pgid);

    if (sgid_cnt>0) {
	str_sgids_cnt = (size_t)sgid_cnt;
	/* str_sgids = malloc (sizeof (int) * str_sgids_cnt); */
	str_sgids = calloc (str_sgids_cnt, sizeof (char **));
	for (i = 0; i < str_sgids_cnt; i++)
	{
	    /* str_sgids[i] = malloc (sizeof (char) * 11); */
	    str_sgids[i] = calloc (11, sizeof (char));
	    snprintf (str_sgids[i], 11, "%u", (unsigned)(sgids[i])); 
	}
	/* transform! - 2^32 uses 10 bytes */
    } else  {
	str_sgids_cnt = 0U;
/*	str_sgids = calloc(1,1);*/
	str_sgids = NULL;
    }





    /*********** B: Obligation UIDGID ***********/
    xacml_obligation_init(&obligation_uidgid,
                          XACML_ATTR_PROFILE_NS_URL_OBLIGATION_UIDGID,
                          XACML_EFFECT_Permit);
    /* uid */
    xacml_obligation_add_attribute(obligation_uidgid,
                                   XACML_ATTR_PROFILE_NS_URL_OBLIGATION_ATTRIBUTES_UID,
                                   XACML_DATATYPE_INTEGER,
                                   str_uid);

    /* gid */
    xacml_obligation_add_attribute(obligation_uidgid,
                                   XACML_ATTR_PROFILE_NS_URL_OBLIGATION_ATTRIBUTES_GID,
                                   XACML_DATATYPE_INTEGER,
                                   str_pgid);

    xacml_response_add_obligation(*response, obligation_uidgid);
    xacml_obligation_destroy(obligation_uidgid);
    /*********** E: Obligation UIDGID ***********/


    /*********** B: Obligation SecondaryGIDs ***********/
    xacml_obligation_init(&obligation_secgids,
                          XACML_ATTR_PROFILE_NS_URL_OBLIGATION_SecondaryGIDs,
                          XACML_EFFECT_Permit);

    for (i = 0; i < str_sgids_cnt; i++)
    {
        xacml_obligation_add_attribute(obligation_secgids,
                                       XACML_ATTR_PROFILE_NS_URL_OBLIGATION_ATTRIBUTES_GID,
                                       XACML_DATATYPE_INTEGER,
                                       str_sgids[i]);
    }

    xacml_response_add_obligation(*response, obligation_secgids);
    xacml_obligation_destroy(obligation_secgids);
    /*********** E: Obligation SecondaryGIDs ***********/



    xacml_response_set_saml_status_code  (*response, SAML_STATUS_Success);
    xacml_response_set_xacml_status_code (*response, XACML_STATUS_ok);
    xacml_response_set_xacml_decision    (*response, XACML_DECISION_Permit);



    /* Free stage */
    free(str_uid);
    free(str_pgid);

    for (i = 0; i < str_sgids_cnt; i++)
    {
        free(str_sgids[i]);
    }
    /* Free if at least one sgid was malloc'ed */
    if (str_sgids_cnt)
        free(str_sgids);
    /* Free stage */ 

    return 0;
}


/************************************************
Function:    printRecvRequest
Description:
             Prints the parsed and fetched data from the XACML request
************************************************/
void printRecvRequest (xacml_profile_request_t * xacml_profile_request)
{
    size_t i;

    scas_log(LOG_DEBUG, "--- Received SAML/XACML Attributes ----------------------------------------------------------\n");
    if (xacml_profile_request -> xacml_profile_subject.xacml_subject_x509_id.subject_x509_id)
        scas_log(LOG_DEBUG, "   %s: %s\n", XACML_ATTR_PROFILE_NS_URL_SUBJECT_X509_ID, xacml_profile_request -> xacml_profile_subject.xacml_subject_x509_id.subject_x509_id);
    if (xacml_profile_request -> xacml_profile_subject.xacml_subject_x509_issuer.subject_x509_issuer)
        scas_log(LOG_DEBUG, "   %s: %s\n", XACML_ATTR_PROFILE_NS_URL_SUBJECT_X509_ISSUER, xacml_profile_request -> xacml_profile_subject.xacml_subject_x509_issuer.subject_x509_issuer);
    if (xacml_profile_request -> xacml_profile_subject.xacml_subject_certificate_serial_number.certificate_serial_number)
        scas_log(LOG_DEBUG, "   %s: %d\n", XACML_ATTR_PROFILE_NS_URL_SUBJECT_CERTIFICATE_SERIAL_NUMBER, 
                                     xacml_profile_request -> xacml_profile_subject.xacml_subject_certificate_serial_number.certificate_serial_number);
    if (xacml_profile_request -> xacml_profile_subject.xacml_cert_chain.cert_chain)
        scas_log(LOG_DEBUG, "   %s: %s\n", XACML_ATTR_PROFILE_NS_URL_CERT_CHAIN, xacml_profile_request -> xacml_profile_subject.xacml_cert_chain.cert_chain);
    if (xacml_profile_request -> xacml_profile_subject.xacml_subject_primary_voms_fqan.subject_primary_voms_fqan)
        scas_log(LOG_DEBUG, "   %s: %s\n", XACML_ATTR_PROFILE_NS_URL_SUBJECT_VOMS_PRIMARY_FQAN, xacml_profile_request -> xacml_profile_subject.xacml_subject_primary_voms_fqan.subject_primary_voms_fqan);

    for (i = 0; i < xacml_profile_request -> xacml_profile_subject.xacml_ca_serial_number_cnt; i++)
        scas_log(LOG_DEBUG, "   %s: [%lu] %d\n", XACML_ATTR_PROFILE_NS_URL_SUBJECT_CA_SERIAL_NUMBER, (long unsigned)i, xacml_profile_request -> xacml_profile_subject.xacml_ca_serial_number[i].ca_serial_number);
    for (i = 0; i < xacml_profile_request -> xacml_profile_subject.xacml_ca_policy_oid_cnt; i++)
        scas_log(LOG_DEBUG, "   %s: [%lu] %s\n", XACML_ATTR_PROFILE_NS_URL_CA_POLICY_OID, (long unsigned)i, xacml_profile_request -> xacml_profile_subject.xacml_ca_policy_oid[i].ca_policy_oid);
    for (i = 0; i < xacml_profile_request -> xacml_profile_subject.xacml_subject_vo_cnt; i++)
        scas_log(LOG_DEBUG, "   %s: [%lu] %s\n", XACML_ATTR_PROFILE_NS_URL_VO, (long unsigned)i, xacml_profile_request -> xacml_profile_subject.xacml_subject_vo[i].subject_vo);
    for (i = 0; i < xacml_profile_request -> xacml_profile_subject.xacml_voms_signing_subject_cnt; i++)
        scas_log(LOG_DEBUG, "   %s: [%lu] %s\n", XACML_ATTR_PROFILE_NS_URL_VOMS_SIGNING_SUBJECT, (long unsigned)i, xacml_profile_request -> xacml_profile_subject.xacml_voms_signing_subject[i].voms_signing_subject);
    for (i = 0; i < xacml_profile_request -> xacml_profile_subject.xacml_voms_signing_issuer_cnt; i++)
        scas_log(LOG_DEBUG, "   %s: [%lu] %s\n", XACML_ATTR_PROFILE_NS_URL_VOMS_SIGNING_ISSUER, (long unsigned)i, xacml_profile_request -> xacml_profile_subject.xacml_voms_signing_issuer[i].voms_signing_issuer);
    for (i = 0; i < xacml_profile_request -> xacml_profile_subject.xacml_voms_dns_port_cnt; i++)
        scas_log(LOG_DEBUG, "   %s: [%lu] %s\n", XACML_ATTR_PROFILE_NS_URL_VOMS_DNS_PORT, (long unsigned)i, xacml_profile_request -> xacml_profile_subject.xacml_voms_dns_port[i].voms_dns_port);
    for (i = 0; i < xacml_profile_request -> xacml_profile_subject.xacml_subject_voms_fqan_cnt; i++)
        scas_log(LOG_DEBUG, "   %s: [%lu] %s\n", XACML_ATTR_PROFILE_NS_URL_SUBJECT_VOMS_FQAN, (long unsigned)i, xacml_profile_request -> xacml_profile_subject.xacml_subject_voms_fqan[i].subject_voms_fqan);
        /* Multiple elements */

    /* Action */
    if (xacml_profile_request -> xacml_profile_action.xacml_action_job_type == ACTION_QUEUE)
        scas_log(LOG_DEBUG, "   %s: %s\n", XACML_ATTR_PROFILE_NS_URL_ACTION_ACTION_ID, XACML_ATTR_PROFILE_NS_URL_ACTION_ACTION_TYPE_QUEUE);
    else if (xacml_profile_request -> xacml_profile_action.xacml_action_job_type == ACTION_EXECUTE_NOW)
        scas_log(LOG_DEBUG, "   %s: %s\n", XACML_ATTR_PROFILE_NS_URL_ACTION_ACTION_ID, XACML_ATTR_PROFILE_NS_URL_ACTION_ACTION_TYPE_EXEC_NOW);
    else if (xacml_profile_request -> xacml_profile_action.xacml_action_job_type == ACTION_ACCESSFILE)
        scas_log(LOG_DEBUG, "   %s: %s\n", XACML_ATTR_PROFILE_NS_URL_ACTION_ACTION_ID, XACML_ATTR_PROFILE_NS_URL_ACTION_ACTION_TYPE_ACCESS_FILE);

    /* Resource */
    if (xacml_profile_request -> xacml_profile_resource.xacml_resource_node_type == RESOURCE_CE)
        scas_log(LOG_DEBUG, "   %s: %s\n", XACML_ATTR_PROFILE_NS_URL_RESOURCE_RESOURCE_ID, XACML_ATTR_PROFILE_NS_URL_RESOURCE_RESOURCE_TYPE_CE);
    else if (xacml_profile_request -> xacml_profile_resource.xacml_resource_node_type == RESOURCE_SE)
        scas_log(LOG_DEBUG, "   %s: %s\n", XACML_ATTR_PROFILE_NS_URL_RESOURCE_RESOURCE_ID, XACML_ATTR_PROFILE_NS_URL_RESOURCE_RESOURCE_TYPE_SE);
    else if (xacml_profile_request -> xacml_profile_resource.xacml_resource_node_type == RESOURCE_WN)
        scas_log(LOG_DEBUG, "   %s: %s\n", XACML_ATTR_PROFILE_NS_URL_RESOURCE_RESOURCE_ID, XACML_ATTR_PROFILE_NS_URL_RESOURCE_RESOURCE_TYPE_WN);
    else if (xacml_profile_request -> xacml_profile_resource.xacml_resource_node_type == RESOURCE_RB)
        scas_log(LOG_DEBUG, "   %s: %s\n", XACML_ATTR_PROFILE_NS_URL_RESOURCE_RESOURCE_ID, XACML_ATTR_PROFILE_NS_URL_RESOURCE_RESOURCE_TYPE_RB);

    if (xacml_profile_request -> xacml_profile_resource.xacml_resource_x509_id.resource_x509_id)
        scas_log(LOG_DEBUG, "   %s: %s\n", XACML_ATTR_PROFILE_NS_URL_RESOURCE_RESOURCE_X509_ID, xacml_profile_request -> xacml_profile_resource.xacml_resource_x509_id.resource_x509_id);
    if (xacml_profile_request -> xacml_profile_resource.xacml_resource_x509_issuer.resource_x509_issuer)
        scas_log(LOG_DEBUG, "   %s: %s\n", XACML_ATTR_PROFILE_NS_URL_RESOURCE_RESOURCE_X509_ISSUER, xacml_profile_request -> xacml_profile_resource.xacml_resource_x509_issuer.resource_x509_issuer);
    if (xacml_profile_request -> xacml_profile_resource.xacml_resource_dns_name.resource_dns_name)
        scas_log(LOG_DEBUG, "   %s: %s\n", XACML_ATTR_PROFILE_NS_URL_RESOURCE_DNS_HOST_NAME, xacml_profile_request -> xacml_profile_resource.xacml_resource_dns_name.resource_dns_name);

    /* Environment */
    /* Todo */
    scas_log(LOG_DEBUG, "--- Received SAML/XACML Attributes ----------------------------------------------------------\n");
}


/************************************************
Function:    xacml_authorize
Description:
             This is the function that is called to get the request and construct the response.
             All the elements from the request are made available by the saml2-xacml2 library and
             it allows to call out to an external facility to get more information that would 
             need to go into the response.
Note:        This is the function that is registered in the SAML2-XACML2 library
             from my server.c example file
Return:


************************************************/
int
xacml_authorize(
    void *                              handler_arg,
    const xacml_request_t               request,
    xacml_response_t                    response)
{
    const char * logstr = "xacml_authorize()";
    int rc = 0;
    int i = 0;
    /* XACML Interoperability Profile Request */
    xacml_profile_request_t * xacml_profile_request;
    lcmaps_account_info_t paccount_info;

#ifdef HAVE_XACML_REQUEST_GET_ID
    const char *req_id=NULL, *resp_id=NULL, *issuer=NULL;
    time_t issue_time;
    char issue_time_str[TIMESTRINGLEN] = "";

    /* Get request ID, issuer and time */
    xacml_request_get_id(request,&req_id);
    xacml_request_get_issuer(request,&issuer);
    xacml_request_get_issue_instant(request,&issue_time);
    if (issue_time>0)
	time_t_to_gmtime_str(issue_time, issue_time_str, TIMESTRINGLEN);

    scas_log(LOG_NOTICE,
	    "%s: Incoming request: %s (issuer %s, issue-time %s)\n",
	    __func__,
	    (req_id ? req_id : "unknown"), (issuer ? issuer : "unknown"),
	    (strlen(issue_time_str) ? issue_time_str : "unknown"));
#endif /* HAVE_XACML_REQUEST_GET_ID */
    
    lcmaps_account_info_init (&paccount_info);

    xacml_profile_request = calloc (1, sizeof (xacml_profile_request_t));
    if (!xacml_profile_request)
    {
        scas_log(LOG_ERR,
		"%s: Out of memory. Couldn't allocate %lu bytes of memory\n",
		logstr, (long unsigned)sizeof (xacml_profile_request_t));
        return -1;
    }
    /* XACML Interoperability Profile Request */



    /* Extract information and push it to the internal structs for direct access
     * to the message elements */
    if (extractRequestSubject (request, xacml_profile_request) < 0)
    {
        scas_log(LOG_ERR, "%s: Failure in extractRequestSubject()\n", logstr);
        goto post_everything_cleanup;
    }

    if (extractRequestAction (request, xacml_profile_request) < 0)
    {
        scas_log(LOG_ERR, "%s: Failure in extractRequestAction()\n", logstr);
        goto post_everything_cleanup;
    }

    if (extractRequestResource (request, xacml_profile_request) < 0)
    {
        scas_log(LOG_ERR, "%s: Failure in extractRequestResource()\n", logstr);
        goto post_everything_cleanup;
    }
    /* Extract information and push it to the internal structs for direct access
     * to the message elements */

    printRecvRequest (xacml_profile_request);   


    /* Now that we will produce a response, set the values */
#ifdef HAVE_XACML_REQUEST_GET_ID
    /* Create and set the ID for the response */
    resp_id = xacml_response_set_id(response, NULL);
#endif /* HAVE_XACML_REQUEST_GET_ID */

    /* Set the XACML response issuer to the pre-determined value */
    xacml_response_set_issuer(response, response_issuer);


#if 0
    /* Disabled LCAS ban0ing on the pilot job subject for the moment */
    /* FIX LCAS ! */
    if ((rc = execLCAS (xacml_profile_request)))
    {
        scas_log(LOG_ERR, "%s: LCAS Authorization failed!\n", logstr);

        /* Construct failure response */
        constructFailureResponse (&response);

        goto post_everything_cleanup;
    } 
#endif


    /*********** Apply mutex ***********/
    pthread_mutex_lock(&mutex);
    /*********** Apply mutex ***********/

    /* use_cert_chain is global static which can be set using
     * set_use_cert_chain() */
    rc = execLCMAPS (xacml_profile_request, use_cert_chain, &paccount_info);
    if (rc)
    {
        scas_log(LOG_ERR, "%s: LCMAPS failed with the return code: %d\n", logstr, rc);

        /*********** Release Mutex ***********/
        pthread_mutex_unlock(&mutex);
        /*********** Release Mutex ***********/

        /* Construct failure response */
        constructFailureResponse (&response);
#ifdef HAVE_XACML_REQUEST_GET_ID
	scas_log(LOG_ERR,
		"%s: SCAS failed, "
		"failure response %s to request %s constructed\n",
		__func__,
		(resp_id ? resp_id : "unknown"), (req_id ? req_id : "unknown"));
#else
	scas_log(LOG_ERR,
		"%s: SCAS failed, failure response constructed\n",
		__func__);
#endif /* HAVE_XACML_REQUEST_GET_ID */

        lcmaps_account_info_clean(&paccount_info);
        goto post_everything_cleanup;
    }

    /*********** Release Mutex ***********/
    pthread_mutex_unlock(&mutex);
    /*********** Release Mutex ***********/


    /* Construct log message of LCMAPS's mapping results */
    scas_log(LOG_DEBUG, "%s: LCMAPS returned: UID: '%d' GID: '%d'\n", 
                       logstr, 
                       paccount_info.uid, 
                       paccount_info.pgid_list[0]);
    for (i = 0; i < paccount_info.nsgid; i++)
        scas_log(LOG_DEBUG, "%s:                             SGID: '%d'\n", 
                           logstr, 
                           paccount_info.sgid_list[i]); 

    /* Construct full response, build success statement in response */
    constructResponse (&response, 
                       paccount_info.uid, 
                       paccount_info.pgid_list[0], 
                       paccount_info.sgid_list, paccount_info.nsgid);

#ifdef HAVE_XACML_REQUEST_GET_ID
    scas_log(LOG_NOTICE,
	    "%s: SCAS succeeded, "
	    "response %s to request %s constructed\n",
	     __func__,
	     (resp_id ? resp_id : "unknown"), (req_id ? req_id : "unknown"));
#else
    scas_log(LOG_NOTICE, "%s: SCAS succeeded, response constructed\n", __func__);
#endif /* HAVE_XACML_REQUEST_GET_ID */

    lcmaps_account_info_clean(&paccount_info);


post_everything_cleanup:
    /* Destroy and free Profile Request */
    destroyRequestSubject (xacml_profile_request);
    destroyRequestAction (xacml_profile_request);
    destroyRequestResource (xacml_profile_request);
    free(xacml_profile_request);

    /* Always '0' */
    return 0;

}
/********** xacml_authorize() **********/

/**
 * Determines the issuer, of the form https://<full canonical hostname>:<port>/
 * When the (first) full canonical name cannot be determined we use the short
 * name, when that also fails, we set the complete issuer to "SCAS" and -1 is
 * returned. This function should be called once before we really start.
 * Note: this code is copy/paste from the EES.
 */
int set_xacml_response_issuer(int portnumber) {
    char tmp_hostname[HOST_NAME_MAX+1], *host;
    struct addrinfo hints, *info;
    int retval;
    int rc=0;	/* return code of this function */

    /* First get (short) hostname */
    if (gethostname(tmp_hostname, HOST_NAME_MAX) != 0)	{
	scas_log(LOG_WARNING,
		"%s: Cannot determine hostname, "
		"will use \"SCAS\" as issuer: %s\n",
		__func__, strerror(errno));
	return -1;
    }
    
    /* Set up hints struct for getting host info */
    memset(&hints, 0, sizeof hints);
    hints.ai_family = AF_UNSPEC;	/* either IPV4 or IPV6 */
    hints.ai_socktype = SOCK_STREAM;
    hints.ai_flags = AI_CANONNAME;

    /* Get info and check it's valid */
    if ( (retval = getaddrinfo(tmp_hostname, NULL, &hints, &info)) != 0 ||
	 info==NULL || info->ai_canonname==NULL ||
	 strlen(info->ai_canonname)>HOST_NAME_MAX ) {
	scas_log(LOG_WARNING,
		"%s: Getting valid hostname failed, will just use shortname: %s\n",
		__func__,
		retval==0 ? "no (valid) info returned" :
		    (retval==EAI_SYSTEM ? strerror(errno) :
					  gai_strerror(retval)));
	host=tmp_hostname;
    } else  {
	/* Found at least one valid entry */
	if (info->ai_next && info->ai_next->ai_canonname)
	    scas_log(LOG_WARNING, "%s: Multiple hostnames found, using first\n",
		    __func__);
	host=info->ai_canonname;
    }

    /* Now try to write the issuer into the global static variable */
    retval = snprintf(response_issuer, ISSUER_MAX,
		      "https://%s:%d/", host, portnumber);
    if (retval<0 || retval>=ISSUER_MAX) {
	scas_log(LOG_WARNING,
		"%s: Cannot set issuer, using \"SCAS\" instead: %s\n",
		__func__, retval<0 ? strerror(errno) : "hostname too long");
	/* Revert to SCAS */
	strcpy(response_issuer,"SCAS");
	rc=-1;
    }

    /* Free address info */
    freeaddrinfo(info);

    /* log on debug */
    scas_log(LOG_DEBUG, "%s: will use \"%s\" as issuer string\n",
	    __func__, response_issuer);

    return rc;
}


/**
 * Set the use_cert_chain global static variable iff set!=-1, return new or
 * existing value.
 */
int set_use_cert_chain(int set)	{
    if (set!=-1)
	use_cert_chain=set;
    return use_cert_chain;
}

/************************************************************************
 * Private functions
 ************************************************************************/

#ifdef HAVE_XACML_REQUEST_GET_ID
static int time_t_to_gmtime_str(const time_t t, char *buf, size_t buf_len)  {
    struct tm res_tm;
    int rc;

    if (gmtime_r(&t, &res_tm)==NULL)
	return -1;

    rc=snprintf(buf, buf_len, "%04d-%02d-%02dT%02d:%02d:%02dZ",
		res_tm.tm_year + 1900,
		res_tm.tm_mon + 1,
		res_tm.tm_mday,
		res_tm.tm_hour,
		res_tm.tm_min,
		res_tm.tm_sec);

    if (rc<0 || (size_t)rc>=buf_len)	{
	buf[0]='\0';
	return -1;
    }

    return 0;
}
#endif /* HAVE_XACML_REQUEST_GET_ID */
